import { compose, merge, createSystem } from 'moulinette';
import { __rest, __assign } from 'tslib';

var rxRules = /^(\&|\@|\:|\#|\.)/;
function isCSS(_, key) {
    if (rxRules.test(key))
        return true;
    return typeof document === 'undefined'
        ? require('known-css-properties').all.includes(require('kebab-case')(key))
        : key in document.body.style;
}
function isEmpty(value) {
    return value === null || (typeof value === 'object' && Object.keys(value).length === 0); // prettier-ignore
}
function isObject(value) {
    return value !== null && typeof value === 'object' && value.constructor === Object; // prettier-ignore
}
function partition(by) {
    return function (props) {
        var match = {};
        var rest = {};
        Object.keys(props).forEach(function (key) {
            var value = props[key];
            if (by(value, key, props)) {
                match[key] = value;
            }
            else {
                rest[key] = value;
            }
        });
        return [match, rest];
    };
}
function compact(props) {
    return partition(isEmpty)(props)[1];
}
function replace(by) {
    var _replace = function (props, context) {
        if (context === void 0) { context = props; }
        var result = {};
        for (var key in props) {
            var parsed = by(props[key], key, context);
            var value = typeof parsed === 'undefined' ? props[key] : parsed;
            result[key] = isObject(value) ? _replace(value, context) : value;
        }
        return result;
    };
    return _replace;
}
function rename(aliases) {
    var _rename = function (props) {
        var renamed = {};
        Object.keys(props).forEach(function (key) {
            var value = props[key];
            var alias = key in aliases ? aliases[key] : key;
            var renamedValue = isObject(value) ? _rename(value) : value;
            if (Array.isArray(alias)) {
                alias.forEach(function (aliasKey) { return (renamed[aliasKey] = renamedValue); });
            }
            else {
                renamed[alias] = renamedValue;
            }
        });
        return renamed;
    };
    return _rename;
}

var aliases = rename({
    bg: 'backgroundColor',
    m: 'margin',
    mx: ['marginLeft', 'marginRight'],
    my: ['marginTop', 'marginBottom'],
    mt: 'marginTop',
    mr: 'marginRight',
    mb: 'marginBottom',
    ml: 'marginLeft',
    p: 'padding',
    px: ['paddingLeft', 'paddingRight'],
    py: ['paddingTop', 'paddingBottom'],
    pt: 'paddingTop',
    pr: 'paddingRight',
    pb: 'paddingBottom',
    pl: 'paddingLeft',
    b: 'border',
    bx: ['borderLeft', 'borderRight'],
    by: ['borderTop', 'borderBottom'],
    bt: 'borderTop',
    br: 'borderRight',
    bb: 'borderBottom',
    bl: 'borderLeft',
    font: 'fontFamily',
    size: 'fontSize'
});

function theme(props, category, key) {
    if (!('theme' in props) || !(category in props.theme))
        return key;
    var values = props.theme[category];
    if (typeof key !== 'undefined') {
        return key in values ? values[key] : key;
    }
    else {
        return values;
    }
}
var rxSpacing = /margin|padding/;
var rxColor = /color|Color/;
var replaceCSS = function (parser) {
    return replace(function (value, key, props) {
        return isCSS(value, key) ? parser(value, key, props) : undefined;
    });
};
var spacing = replaceCSS(function (value, key, props) {
    if (typeof value === 'number' && rxSpacing.test(key)) {
        return value * (theme(props, 'spacing') || 1);
    }
});
var fonts = replaceCSS(function (value, key, props) {
    if (key === 'fontFamily') {
        return theme(props, 'fonts', value);
    }
});
var sizes = replaceCSS(function (value, key, props) {
    if (key === 'fontSize') {
        return theme(props, 'sizes', value);
    }
});
var colors = replaceCSS(function (value, key, props) {
    if (key === 'background' || rxColor.test(key)) {
        return theme(props, 'colors', value);
    }
});
var applyTheme = compose([spacing, fonts, sizes, colors]);

function isNumber(key) {
    return typeof key === 'number' || !isNaN(Number(key));
}
function isResponsive(breakpoints) {
    return function (value, key) {
        return isObject(value) && isObject(breakpoints) && key in breakpoints;
    };
}
function mediaQuery(key, breakpoints) {
    if (breakpoints === void 0) { breakpoints = {}; }
    var value = key in breakpoints ? breakpoints[key] : key;
    if (typeof value === 'string' && value.startsWith('@media')) {
        return value;
    }
    var breakpoint = isNumber(value) ? value + 'px' : value;
    return "@media screen and (min-width: " + breakpoint + ")";
}
function responsive(input) {
    var breakpoints = theme(input, 'breakpoints');
    var _a = partition(isResponsive(breakpoints))(input), responsive = _a[0], props = _a[1];
    var mediaQueries = Object.keys(responsive).map(function (key) {
        var _a;
        return (_a = {},
            _a[mediaQuery(key, breakpoints)] = responsive[key],
            _a);
    });
    return mediaQueries.reduce(merge, props);
}

var partitionCSS = partition(isCSS);
function extractCSS(_a) {
    var _b = _a.css, nextCSS = _b === void 0 ? [] : _b, input = __rest(_a, ["css"]);
    var _c = partitionCSS(input), cssProps = _c[0], props = _c[1];
    var computedCSS = compact(cssProps);
    var css = isEmpty(computedCSS) ? nextCSS : [computedCSS].concat(nextCSS);
    return css.length > 0 ? __assign({}, props, { css: css }) : props;
}
var compileCSS = compose([extractCSS, applyTheme, responsive, aliases]);

function stringify(strings, args, props) {
    var computed = args.map(function (arg) {
        return typeof arg === 'function' ? arg(props) : arg;
    });
    return strings.reduce(function (template, string, i) {
        return template + string + (i in computed ? computed[i] : '');
    }, '');
}
function addTemplate(strings, args) {
    return function (input) {
        var _a = compileCSS(input) || {}, _b = _a.css, css = _b === void 0 ? [] : _b, props = __rest(_a, ["css"]);
        var str = stringify(strings, args, props);
        return __assign({}, props, { css: [str].concat(css) });
    };
}

function createElement(_a) {
    var _b = _a.as, type = _b === void 0 ? 'div' : _b, props = __rest(_a, ["as"]);
    return __assign({}, props, { as: type });
}

function addAnimation(nextAnimation) {
    return function (_a) {
        var _b = _a.animation, animation = _b === void 0 ? '' : _b, _c = _a.animations, animations = _c === void 0 ? [] : _c, props = __rest(_a, ["animation", "animations"]);
        return (__assign({}, props, { animations: animations.concat([nextAnimation.trim(), animation]).filter(Boolean) }));
    };
}
function combineAnimations(_a) {
    var _b = _a.animations, animations = _b === void 0 ? [] : _b, props = __rest(_a, ["animations"]);
    return animations.length > 0
        ? __assign({}, props, { animation: animations.join() }) : props;
}

var systyle = compose([createElement, compileCSS, combineAnimations]);
function mixin(Styled) {
    Styled.as = function (type) { return Styled.with({ as: type }); };
    Styled.css = function (strings) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        return Styled.with(addTemplate(strings, args));
    };
    Styled.animate = function (duration, animation) {
        if (animation === void 0) { animation = ''; }
        return Styled.with(addAnimation(animation + " " + duration));
    };
}
function createStyled(builder) {
    return createSystem(builder)
        .extend(mixin)
        .with(systyle);
}

export default createStyled;
export { systyle, mixin, createStyled };
